#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2023 "Dark and Darker Epic" May 11, 2023

in vec3 in_Position;
out vec4 uv;
out float time;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};
layout(std140) uniform MidiBlock {
	vec4 midi[64];//x = note id, y = velocity, z = channel/track, w = time in seconds since note pressed
	int midiCount;
};
out vec2 music[6];
/*LMMS midi converter https://www.lynxwave.com/LMMStoMIDI/LMMStoMIDIConverter.html since LMMS midi export is broken.

Launch options for midi config 'Shaderview console DarkAndDarkerEpic.mid 3 1'
3 is the bitmask 0b11 signifying which instruments are beats/drums with a fixed length.
1 is the fixed beat length in seconds.

Midi Instrument 0 = Strum Guitar(drum)
Midi Instrument 1 = Oud(drum)
Midi Instrument 2 = Strings
Midi Instrument 3 = Choir*/


#define SAMPLES 10000000
//#define ANIMATION 1

#define EPSILON .004
#define NRM_EPSILON (EPSILON*.5)

#define FOG(rp) vec3(0)
#define FOG_DENSITY 0.05

#define PI 3.141592653589793


//2d rotation matrix
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

//pseudo random uvec4 hash
uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))

//r2 hash http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
vec2 quasiRandom(float n) {
	#define R2V 1.32471795724474602596
	return fract(.5+n*(1./vec2(R2V,R2V*R2V)));
}



//geometry distance/material function
vec2 geometry(vec3 rp) {
	bool axis;
	vec2 geo;
	vec2 ax = vec2(5.,10.)-abs(rp.xy-vec2(0,5.));//walls
	if (ax.x < ax.y) {
		geo.x = ax.x;
		axis = false;
	} else {
		geo.x = ax.y;
		axis = true;
	}
	geo.y = 0.;

	if (geo.x < .4) {
		//rock/cobble displacement on walls
		float deform = 0.;
		if (axis) ax = rp.xz+sign(rp.y)*.92783;
		else ax = rp.yz+sign(rp.x)*2.287234;
		for (float l = 1.; l < 4.; l++) {
			vec2 ap = ax+cos(ax.yx*l+vec2(3.782384,32.784235)*l*l)*.3/l;
			ap = fract(abs(ap)*l)-.5;
			deform += max(0.,1.-length(ap)*2.)*.2/l;
		}
		if (rp.y < 0.) deform *= min(1.,abs(rp.x+cos(rp.z*.73764+2.*cos(rp.z*.55)))/3.);
		geo.x -= deform;
	}
	
	//torches
	float rz = mod(abs(rp.z),8.)-4.;
	vec3 tp = vec3(abs(rp.x),rp.y,rz)-vec3(4.8,.6,0);
	//torch culling bounds
	float tb = max(length(tp.xz)-.7,abs(tp.y-1.4)-1.8);
	if (tb < .5) {
		//torch metal holder
		float td = max(length(tp)-.4,tp.y+.06);
		if (td < geo.x) {
			geo = vec2(td,1);
		}
	
		//torch flame
		float ylim = tp.y+.1,
			tx = step(rp.x,0.), tz = floor(rp.z/8.+1000.),
			ht = .2+music[int(tz)%5][int(tx)];
		tp.x += .1+cos(rp.z*16.5)*.2;
		tp.y -= .4+ht*.9+cos(rp.x*12.+2.16328-time)*.2;//+(cos(rp.x*12.+2.17823)+cos(rp.y*28.-time*10.)+cos(rp.z*16.5))*.2;
		tp.z += cos(rp.y*13.-time*20.)*.2;
		tp.y = max(abs(tp.y)-ht,0.);
		float fd = max(length(tp)-.4,-ylim);
		if (fd < geo.x) {
			geo = vec2(fd,2);
		}
		
	} else if (tb < geo.x) {
		geo = vec2(tb,1);
	}
	
	return geo;
}

//returns mat4x3(vec3 diffuse, vec3 specular, vec3 specularData(x=amount, y=roughness, z=fresnel), vec3 emission)
mat4x3 material(vec3 rp, vec3 nrm, float mat) {
	if (mat == 0.) {
		//cobble/rock wall
		return mat4x3(vec3(.3,.25,.2),vec3(.15,.13,.12),vec3(.3,.14,.7),vec3(0));
	}
	
	if (mat == 1.) {
		//torch holder metal
		return mat4x3(vec3(.05), vec3(.7), vec3(.5,.01,0.8), vec3(0));
	}
	
	//torch flame
	return mat4x3(vec3(.9),vec3(0),vec3(0), mix(vec3(1,.7,.3),vec3(.4,.7,1),music[5].y)*(5.+music[5].x));
}

//calculate normal from geometry distance difference
vec3 normal(vec3 rp) {
	vec3 nrm = vec3(geometry(rp+vec3(NRM_EPSILON,0,0)).x,geometry(rp+vec3(0,NRM_EPSILON,0)).x,geometry(rp+vec3(0,0,NRM_EPSILON)).x)-geometry(rp).x;
	return nrm==vec3(0) ? vec3(0,1,0) : normalize(nrm);
}

//return background color from ray
vec3 background(vec3 rp, vec3 rd) {
	return vec3(0);
}
vec3 backgroundLighting(vec3 rp, vec3 rd) {
	return background(rp,rd);
}



uniform sampler2D Buffer0;

void vert() {

	#ifdef ANIMATION
	time = float(frame)/float(framerate*SAMPLES);
	
	float strv = 7.*clamp(time/15.-1.,0.,1.);
	for (int i = 0; i < 6; i++) music[i] = vec2(0);
	for (int i = 0; i < midiCount; i++) {
		float ch = midi[i].z;
		if (ch == 0.) {
			//guitar
			music[int(midi[i].x)%5].x += pow(max(0.,1.-midi[i].w)*min(1.,midi[i].w*10.),2.)*1.4;
		
		} else if (ch == 1.) {
			//oud
			music[int(midi[i].x)%5].y += pow(max(0.,1.-midi[i].w)*min(1.,midi[i].w*10.),2.)*1.4;
			
		} else if (ch == 2.) {
			//strings
			music[5].x += strv;
		} else if (ch == 3.) {
			//choir
			music[5].y++;
		}
	}
	#else
	time = 30.;
	music[5].x = 7.;
	
	//time = 0.;
	//for (int i = 0; i < 5; i++) music[i] = cos(float(i+1)*vec2(.8,2.5))*.5+.5;
	//music[5] = vec2(7,1);
	#endif
	
	//anti aliasing jitter
	uv = vec4((in_Position.xy*vec2(screenX/screenY,1)+(quasiRandom(float(frame%SAMPLES))-.5)*2./screenY),
		(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera origin position and direction
	vec3 rp = vec3(0,-1,2+time*.34),
		rd = normalize(vec3(uv.xy,1.5));

	//path tracing
	bool hit = false;
	vec3 color = vec3(1), light = vec3(0);
	#ifdef FOG
	vec3 initRp = rp, hitRp;
	#endif
	rp += rd*4.;
	for (int i = 0; i < 80; i++) {
		vec2 geo = geometry(rp);
		if (geo.x < EPSILON) {
			//hit get object/material data
			vec3 nrm = normal(rp);
			mat4x3 mat = material(rp,nrm,geo.y);
			vec4 rnd = hash(uvec4(abs(rp)*10000.,frame))*I2F;
			
			//fog
			#ifdef FOG
			if (!hit) hitRp = rp;
			#endif
			hit = true;
			
			//object emission
			light += color*mat[3];
			
			//choose diffuse/specular ray
			if (rnd.w >= mat[2].x) {
				//diffuse(uniformly distributed ray direction around hemisphere of normal)
				color *= mat[0];
				
				float cs = cos(rnd.x*PI), sn = sqrt(1.-cs*cs), ang = rnd.y*PI*2.;
				rd = normalize(vec3(cs,sn*cos(ang),sn*sin(ang)));//random ray direction
				if (dot(rd,nrm) < 0.) rd = - rd;//limit bounce to normal hemisphere
			
			} else {
				//specular(reflection off inverse distributed direction)
				float fresnel = clamp((1.+dot(rd,nrm)-(1.-mat[2].z))/max(1e-6,mat[2].z),0.,1.),
					fresi = 1.-fresnel;
				color *= mat[1]*fresi+fresnel;
				
				vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);//tangent and binormal
				float cs = pow(rnd.x,mat[2].y*max(0.,fresi*2.-1.)), sn = sqrt(1.-cs*cs), ang = rnd.y*PI*2.;
				rd = normalize(reflect(rd, mat3(nrm,tang,bitang)*normalize(vec3(cs,sn*cos(ang),sn*sin(ang)))));//reflect off random direction(simulate microsurface bumps)
			}
			
			//step ray back out of objects
			rp += nrm*(EPSILON*2.-geo.x);
			
		} else {
			//step ray forward
			rp += rd*geo.x;
		}		
	}
	
	//apply background
	light += hit ? color*backgroundLighting(rp,rd) : background(rp,rd);
	
	//apply fog
	#ifdef FOG
	float fi = 1./exp(length(hitRp-initRp)*FOG_DENSITY);
	light = light*fi+FOG(hitRp)*(1.-fi);
	#endif
	
	//add sample to buffer
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (frame%SAMPLES == 0) sum = vec4(0);
	sum += vec4(light, 1.);
	gl_FragColor = sum;
}
